const SHA256 = require('crypto-js/sha256');


/*=== Generiert einen Block ===*/
class Block {
    constructor(index, timestamp, data, previousHash =''){
        this.index= index;
        this.timestamp = timestamp;
        this.data = data;
        this.previousHash = previousHash;
        this.hash = this.calculateHash();
        this.nonce = 0;
    }

    /* ===Generierung eines zufälligen Hashwertes=== */
    calculateHash() {
        return SHA256(this.index + this.previousHash + this.timestamp + JSON.stringify(this.data) + this.nonce).toString();
    }

    /*Festlegung der Geschwindigkeit wie schnell ein Bock gemined werden kann
    * Änderung des this.difficulty Wert verlängert die Zeit bis ein Block gemined wird
    * bsp. Bitcoin braucht 10 Minuten*/
    mineBlock(difficulty) {
        while(this.hash.substring(0,difficulty) !== Array(difficulty + 1 ).join("0")){
            this.nonce++;
            this.hash = this.calculateHash();
        }
        console.log("Block mined: " + this.hash);
    }
}

/*=== erstellt die Blockchain ===*/
class Blockchain {
    constructor(){
        this.chain = [this.createGenesisBlock()];
        //wie lange dauert es bis ein neuer block geschrieben werden darf
        this.difficulty = 1;
    }

    /* ===Erster Block in einer Blockchain=== */
    createGenesisBlock(){
        return new Block(0, "25/01/2019", "Genesis block" , "0")
    }

    getLatestBlock(){
        return this.chain[this.chain.length - 1];
    }
    addBlock(newBlock){
        newBlock.previousHash = this.getLatestBlock().hash;
        newBlock.mineBlock(this.difficulty);
        //newBlock.hash = newBlock.calculateHash();
        this.chain.push(newBlock);
    }

    /*=== hält der previousBlock den Hash des currentBlock ===*/
    isChainValid() {
        for(let i = 1; i < this.chain.length; i++){
            const currentBlock = this.chain[i];
            const previousBlock = this.chain[i - 1];

            if(currentBlock.hash !== currentBlock.calculateHash()){
                return false;
            }
            if(currentBlock.previousHash !== previousBlock.hash){
                return false;
            }
        }
        return true;
    }
}



/*=== Test Area ===*/
let toastCoin = new Blockchain();

/* ===Erstellung von zwei Blöcken=== */
console.log("Mining block 1...");
toastCoin.addBlock(new Block(1, "25/01/2019", {amount:4}));
console.log("Mining block 2...")
toastCoin.addBlock(new Block(2, "25/01/2019", {amount:10}));

/*Muss true sein, weil die Hashes übereinstimme*/
console.log("is Blockchain valid: " + toastCoin.isChainValid());

/*Block wird manipuliert*/
/*Versuch dem ersten block von 4 auf 100 zu ändern*/
toastCoin.chain[1].data = { amount: 100};
console.log("is Blockchain valid: " + toastCoin.isChainValid());

/*Verusch den Hashwert des ersten blockes zu ändern*/
toastCoin.chain[1].hash = toastCoin.chain[1].calculateHash();
console.log("is Blockchain valid: " + toastCoin.isChainValid());


//
// console.log(JSON.stringify(toastCoin, null, 4));
